ICE.eigen.L <- function(K) {
  eig <- eigen(K,symmetric=TRUE)
  return(list(values=eig$values,vectors=eig$vectors))
}

ICE.eigen.R <- function(K, X) {
  n <- nrow(K)
  q <- ncol(X)

  S <- diag(1,n)-X%*%solve(crossprod(X,X))%*%t(X)
  eig <- eigen(S%*%(K+diag(1,n))%*%t(S),symmetric=TRUE)

  if ( is.complex(eig$values) ) {
    eig$values <- Re(eig$values)
    eig$vectors <- Re(eig$vectors)
  }
  return(list(values=eig$values[1:(n-q)]-1,vectors=eig$vectors[,1:(n-q)]))
}

# log-likelihood of maximum likelihood (internal)
ICE.ML.LL <- function(logdelta, lambda, etas, xi) {
  n <- length(xi)
#  print(xi)
#  print(etas)
  delta <- exp(logdelta)
  return( 0.5*(n*(log(n/(2*pi))-1-log(sum((etas*etas)/(lambda+delta))))-sum(log(xi+delta))) )
}

# derivative of log-likelihood function (internal)
ICE.ML.dLL <- function(logdelta, lambda, etas, xi) {
  n <- length(xi)
  delta <- exp(logdelta)
  etasq <- etas*etas
  ldelta <- lambda+delta
  return( 0.5*(n*sum(etasq/(ldelta*ldelta))/sum(etasq/ldelta)-sum(1/(xi+delta))) )
}

ICE.REML.LL <- function(logdelta, lambda, etas) {
  nq <- length(etas)
  delta <-  exp(logdelta)
  return( 0.5*(nq*(log(nq/(2*pi))-1-log(sum(etas*etas/(lambda+delta))))-sum(log(lambda+delta))) )
}

ICE.REML.dLL <- function(logdelta, lambda, etas) {
  nq <- length(etas)
  delta <- exp(logdelta)
  etasq <- etas*etas
  ldelta <- lambda+delta
  return( 0.5*(nq*sum(etasq/(ldelta*ldelta))/sum(etasq/ldelta)-sum(1/ldelta)) )
}

# global MLE optimization with grid search and Newton-Raphson algorithm
ICE.MLE.grid <- function(y, K, X, nstep=100, llim=-10, ulim=10, esp=1e-10, eigen.L = NULL, eigen.R = NULL) {
  n <- length(y)        # number of individuals
  q <- ncol(X)

  stopifnot(nrow(K) == n)
  stopifnot(ncol(K) == n)

  # Immediately return if X is singular.
  # If SNPs are all identical, X) becomes singular
  if ( det(crossprod(X,X)) == 0 ) 
    {
      warning("X is singular")
      return (list(delta=0,LL=0,ve=0,va=0))
    }

  if ( is.null(eigen.L) ) {
    eigen.L <- ICE.eigen.L(K)
  }

  if ( is.null(eigen.R) ) {
    eigen.R <- ICE.eigen.R(K,X)
  }

  etas <- crossprod(eigen.R$vectors,y)
  
  logdelta <- (0:nstep)/nstep*(ulim-llim)+llim
  m <- length(logdelta)
  delta <- exp(logdelta)
  Lambdas <- matrix(eigen.R$values,n-q,m) + matrix(delta,n-q,m,byrow=TRUE)
  Xis <- matrix(eigen.L$values,n,m) + matrix(delta,n,m,byrow=TRUE)
  Etasq <- matrix(etas*etas,n-q,m)
  LL <- 0.5*(n*(log(n/(2*pi))-1-log(colSums(Etasq/Lambdas)))-colSums(log(Xis)))
  dLL <- 0.5*delta*(n*colSums(Etasq/(Lambdas*Lambdas))/colSums(Etasq/Lambdas)-colSums(1/Xis))

  optlogdelta <- vector(length=0)
  optLL <- vector(length=0)
  if ( dLL[1] < esp ) {
    optlogdelta <- append(optlogdelta, llim)
    optLL <- append(optLL, ICE.ML.LL(llim,eigen.R$values,etas,eigen.L$values))
  }
  if ( dLL[m-1] > 0-esp ) {
    optlogdelta <- append(optlogdelta, ulim)
    optLL <- append(optLL, ICE.ML.LL(ulim,eigen.R$values,etas,eigen.L$values))
  }

  for( i in 1:(m-1) )
    {
      if ( ( dLL[i]*dLL[i+1] < 0 ) && ( dLL[i] > 0 ) && ( dLL[i+1] < 0 ) ) 
#      if ( ( dLL[i]*dLL[i+1] < 0 ) && ( abs(dLL[i]) > esp ) && ( abs(dLL[i+1]) > esp ) ) 
#      if ( dLL[i]*dLL[i+1] < 0-esp )
        {
          r <- uniroot(ICE.ML.dLL, lower=logdelta[i], upper=logdelta[i+1], lambda=eigen.R$values, etas=etas, xi=eigen.L$values)
          optlogdelta <- append(optlogdelta, r$root)
          optLL <- append(optLL, ICE.ML.LL(r$root,eigen.R$values, etas, eigen.L$values))
        }
    }
  optdelta <- exp(optlogdelta)

  maxdelta <- exp(optlogdelta[which.max(optLL)])
  maxLL <- max(optLL)
  maxva <- sum(etas*etas/(eigen.R$values+maxdelta))/n
  maxve <- maxva*maxdelta

  return (list(delta=maxdelta,LL=maxLL,ve=maxve,va=maxva))
}

ICE.REMLE.grid <- function(y, K, X, nstep=100, llim=-10, ulim=10, esp=1e-10, eigen.R = NULL) {
  n <- length(y)        # number of individuals
  q <- ncol(X)

  stopifnot(nrow(K) == n)
  stopifnot(ncol(K) == n)

  # Immediately return if X is singular.
  # If SNPs are all identical, X) becomes singular
  if ( det(crossprod(X,X)) == 0 ) 
    {
      warning("X is singular")
      return (list(delta=0,LL=0,ve=0,va=0))
    }

  if ( is.null(eigen.R) ) {
    eigen.R <- ICE.eigen.R(K, X)
  }

  etas <- crossprod(eigen.R$vectors,y)

  logdelta <- (0:nstep)/nstep*(ulim-llim)+llim
  m <- length(logdelta)
  delta <- exp(logdelta)
  Lambdas <- matrix(eigen.R$values,n-q,m) + matrix(delta,n-q,m,byrow=TRUE)
  Etasq <- matrix(etas*etas,n-q,m)
  LL <- 0.5*((n-q)*(log((n-q)/(2*pi))-1-log(colSums(Etasq/Lambdas)))-colSums(log(Lambdas)))
  dLL <- 0.5*delta*((n-q)*colSums(Etasq/(Lambdas*Lambdas))/colSums(Etasq/Lambdas)-colSums(1/Lambdas))

  optlogdelta <- vector(length=0)
  optLL <- vector(length=0)
  if ( dLL[1] < esp ) {
    optlogdelta <- append(optlogdelta, llim)
    optLL <- append(optLL, ICE.REML.LL(llim,eigen.R$values,etas))
  }
  if ( dLL[m-1] > 0-esp ) {
    optlogdelta <- append(optlogdelta, ulim)
    optLL <- append(optLL, ICE.REML.LL(ulim,eigen.R$values,etas))
  }

  for( i in 1:(m-1) )
    {
        if ( ( dLL[i]*dLL[i+1] < 0 ) && ( dLL[i] > 0 ) && ( dLL[i+1] < 0 ) )
#      if ( ( dLL[i]*dLL[i+1] < 0 ) && ( abs(dLL[i]) > esp ) && ( abs(dLL[i+1]) > esp ) ) 
#      if ( dLL[i]*dLL[i+1] < 0-esp )
        {
          r <- uniroot(ICE.REML.dLL, lower=logdelta[i], upper=logdelta[i+1], lambda=eigen.R$values, etas=etas)
          optlogdelta <- append(optlogdelta, r$root)
          optLL <- append(optLL, ICE.REML.LL(r$root,eigen.R$values, etas))
        }
    }
  optdelta <- exp(optlogdelta)

  maxdelta <- exp(optlogdelta[which.max(optLL)])
  maxLL <- max(optLL)
  maxva <- sum(etas*etas/(eigen.R$values+maxdelta))/(n-q)
  maxve <- maxva*maxdelta

  return (list(delta=maxdelta,LL=maxLL,ve=maxve,va=maxva))
}

ICE.LRT.batch <- function(ys, xs, resonly = FALSE, standardize = TRUE, nstep=100, llim=-10, ulim=10, esp=1e-10) {
  m <- nrow(xs)
  g <- nrow(ys)
  n <- ncol(ys)

  stopifnot(ncol(xs) == n)

  chisqs <- matrix(nrow=g,ncol=m)
  ps <- matrix(nrow=g,ncol=m)
  vas <- matrix(nrow=g,ncol=m)
  ves <- matrix(nrow=g,ncol=m)
  vas0 <- matrix(nrow=g,ncol=m)
  ves0 <- matrix(nrow=g,ncol=m)
  nvs <- vector(length=m)

  y <- ys
  for(i in 1:m) {
    print(i)
    x <- xs[i,]
    vids <- !is.na(x)
    nv <- sum(vids)
    if ( nv < n ) {
      x <- x[vids]
      y <- ys[,vids]
    }

    if ( resonly ) {
      x0 <- x - mean(x)
      M <- tcrossprod(x0,x0)/as.double(crossprod(x0,x0))
      y <- y - matrix(rowMeans(y),g,nv) - crossprod(y%*%M)
    }
    else if ( standardize ) {
      y <- y - matrix(rowMeans(y),g,nv)
    }
    
    if ( standardize ) {
      y <- y/matrix(apply(y,1,sd),g,nv)
      K <- cov(y)+diag(1e-6,nv)
    }
    else {
      K <- cov(y)+diag(1e-6,nv)
    }

    X <- cbind(matrix(1,nv,1),x)

    eig.L <- ICE.eigen.L(K)
    eig.R0 <- ICE.eigen.R(K,matrix(1,nv,1))
    eig.R <- ICE.eigen.R(K,X)
    for ( j in 1:g ) {
      yj <- t(y[j,,drop=FALSE])
      MLE0 <- ICE.MLE.grid(yj, K, matrix(1,nv,1),nstep,llim,ulim,esp,eig.L,eig.R0)
      MLE1 <- ICE.MLE.grid(yj, K, X, nstep, llim, ulim,esp,eig.L, eig.R)
      chisqs[j,i] <- 2*(MLE1$LL-MLE0$LL)
      vas[j,i] <- MLE1$va
      ves[j,i] <- MLE1$ve
      vas0[j,i] <- MLE0$va
      ves0[j,i] <- MLE0$ve
    }
    ps[,i] <- pchisq(chisqs[,i],1,lower.tail=FALSE)
    nvs[i] <- nv
  }
  
  return(list(chisqs=chisqs,ps=ps,vas=vas,ves=ves,vas0=vas0,ves0=ves0,nvs=nvs))
}

ICE.REMLt.batch <- function(ys, xs, resp = 1, standardize = TRUE, nstep=100, llim=-10, ulim=10, esp=1e-10) {
  m <- nrow(xs)
  g <- nrow(ys)
  n <- ncol(ys)

  stopifnot(ncol(xs) == n)

  ts <- matrix(nrow=g,ncol=m)
  ps <- matrix(nrow=g,ncol=m)
  vas <- matrix(nrow=g,ncol=m)
  ves <- matrix(nrow=g,ncol=m)
  nvs <- vector(length=m)
  nrs <- rep(0,m)

  if ( resp == 1 ) {
    yc <- ys - matrix(rowMeans(ys),g,n)
    if ( standardize ) {
      yc <- yc/matrix(apply(yc,1,sd),g,n)
    }
    K0 <- cov(yc)
    eig.L0 <- ICE.eigen.L(K0)
  }

  for(i in 1:m) {
    print(i)
    if ( ( i > 1 ) && ( sum(xs[i-1,] != xs[i,],na.rm=T) == 0 ) ) {
      ts[,i] <- ts[,i-1]
      vas[,i] <- vas[,i-1]
      ves[,i] <- ves[,i-1]
      ps[,i] <- ps[,i-1]
      nvs[i] <- nvs[i-1]
      nrs[i] <- nrs[i-1]
    }
    else {
      x <- xs[i,]
      vids <- !is.na(x)
      nv <- sum(vids)
      if ( nv < n ) {
        x <- x[vids]
        y <- ys[,vids]
      }
      else {
        y <- ys
      }
      
      if ( resp < 1 ) {
        x0 <- x - mean(x)
        yc <- y - matrix(rowMeans(y),g,nv)
        rc <- cor(t(yc),x0)
        pc <- pf(rc*rc*(nv-2)/(1-rc*rc),1,nv-2,lower.tail=FALSE)
        ids <- pc<resp
        nr <- sum(ids)
        if ( g-nr < nv*5 ) {
          ids <- rank(pc,ties.method="random") <= (g-nv*5)
          nr <- sum(ids)
        }
        yc <- yc[!ids,]
                                        #yc[ids,] <- yc[ids,]%*%(diag(1,nv)-x0%*%solve(crossprod(x0,x0))%*%t(x0))
        nrs[i] <- nr
        
        if ( standardize ) {
          yc <- yc/matrix(apply(yc,1,sd),g-nr,nv)
          K <- cov(yc)
        }
        else {
          K <- cov(yc)
        }
        eig.L <- ICE.eigen.L(K)
      }
      else {
        if ( nv == n ) {
          K <- K0
          eig.L <- eig.L0
        }
        else {
          K <- K0[vids,vids]
          eig.L <- ICE.eigen.L(K)
        }
      }
      
      X <- cbind(matrix(1,nv,1),x)
      eig.R <- ICE.eigen.R(K,X)
      
      for ( j in 1:g ) {
        yj <- y[j,]
        REMLE <- ICE.REMLE.grid(yj, K, X, nstep, llim, ulim, esp, eig.R)
        U <- eig.L$vectors%*%diag(sqrt(1/(eig.L$values+REMLE$delta)))
        Xt <- crossprod(U,X)
        yt <- crossprod(U,yj)
        iXX <- solve(crossprod(Xt,Xt))
        beta <- iXX%*%crossprod(Xt,yt)
        ts[j,i] <- beta[2]/sqrt(iXX[2,2]*REMLE$va)
        vas[j,i] <- REMLE$va
        ves[j,i] <- REMLE$ve
      }
      ps[,i] <- pt(abs(ts[,i]),nv-1,lower.tail=FALSE)*2
      nvs[i] <- nv
    }
  }
  return(list(ts=ts,ps=ps,vas=vas,ves=ves,nvs=nvs,nrs=nrs))
}

ICE.simul.exprs <- function(ys, xs, cis.size) {
  m <- nrow(xs)
  g <- nrow(ys)
  n <- ncol(ys)

  stopifnot(ncol(xs) == n)

  ys <- ys - matrix(rowMeans(ys),g,n)
  ys <- ys/matrix(apply(ys,1,sd),g,n)
  K <- cov(ys)
  K <- K/sum(diag(K)/n)
  Kc <- chol(K+diag(1e-10,n))
  sd.xs <- apply(xs,1,sd,na.rm=T)
  zs <- cis.size/sd.xs*xs + matrix(rnorm(m*n),m,n)%*%Kc
  return(zs)
}

ICE.REMLt.diag.batch.nomiss <- function(exprs, ys, xs, standardize = TRUE, nstep=100, llim=-10, ulim=10, esp=1e-10) {
  m <- nrow(xs)
  n <- ncol(ys)

  stopifnot(ncol(xs) == n)
  stopifnot(nrow(ys) == m)


  ts <- vector(length=m)
  ps <- vector(length=m)
  vas <- vector(length=m)
  ves <- vector(length=m)

  yc <- exprs - matrix(rowMeans(exprs),nrow(exprs),n)
  if ( standardize ) {
    yc <- yc/matrix(apply(yc,1,sd),nrow(exprs),n)
  }
  K <- cov(yc)
  eig.L <- ICE.eigen.L(K)

  for(i in 1:m) {
    if ( i %% 1000 == 0 ) {
      print(i)
    }
    x <- xs[i,]
    y <- ys[i,]
    X <- cbind(matrix(1,n,1),x)
    eig.R <- ICE.eigen.R(K,X)
    REMLE <- ICE.REMLE.grid(y, K, X, nstep, llim, ulim, esp, eig.R)
    U <- eig.L$vectors%*%diag(sqrt(1/(eig.L$values+REMLE$delta)))
    Xt <- crossprod(U,X)
    yt <- crossprod(U,y)
    iXX <- solve(crossprod(Xt,Xt))
    beta <- iXX%*%crossprod(Xt,yt)
    ts[i] <- beta[2]/sqrt(iXX[2,2]*REMLE$va)
    ps[i] <- pt(abs(ts[i]),n-1,lower.tail=FALSE)*2
  }
  return(ps)
}

ICE.vartest.batch <- function(exprs, standardize = TRUE, nstep=100, llim=-10, ulim=10, esp=1e-10) {
  g <- nrow(exprs)
  n <- ncol(exprs)

  ps <- vector(length=g)

  y <- exprs
  if ( standardize ) {
      y <- y - matrix(rowMeans(y),g,n)
      y <- y/matrix(apply(y,1,sd),g,n)
      K <- cov(y)+diag(1e-6,n)
  }
  else {
    K <- cov(y)+diag(1e-6,n)
  }

  K0 <- diag(1,n)
  eig.L <- ICE.eigen.L(K)
  eig.R <- ICE.eigen.R(K,matrix(1,n,1))
  eig.L0 <- ICE.eigen.L(K0)
  eig.R0 <- ICE.eigen.R(K0,matrix(1,n,1))

  for ( j in 1:g ) {
    yj <- t(y[j,,drop=FALSE])
    MLE0 <- ICE.MLE.grid(yj, diag(1,n), matrix(1,n,1),nstep,llim,ulim,esp,eig.L0,eig.R0)
    MLE1 <- ICE.MLE.grid(yj, K, matrix(1,n,1),nstep,llim,ulim,esp,eig.L,eig.R)
    ps[j] <- pchisq(2*(MLE1$LL-MLE0$LL),1,lower.tail=FALSE)/2;
  }
  
  return(ps)
}
